
local version = tonumber(string.sub(love._version,1,3))

local kc = 'upupdowndownleftrightleftrightba'
local ck = 1

local tileset
local tileWidth = 16
local tileHeight = 16
local tileQuads = {}
local tileBatch = {}
local speed = 1

local width,length,height = math.floor(800/tileWidth),math.floor(600/tileHeight),5

local left,front,right,back,up,down = 1,2,4,8,16,32

local zoom = 1

local map = {}
local path = {}

local function setCaption(caption)
	if version >= 0.9 then
		love.window.setCaption(caption)
	else
		love.graphics.setCaption(caption)
	end
end

local function updateCaption()
	setCaption('Layer:' .. zoom .. ',speed=' .. speed)
end

local function lerp(a,b,t)
	return a + (b - a) * t
end

local function delerp(a,b,c)
	return (c - a)/(b - a)
end

local function setMaze(x,y,z,v)
	if version >= 0.9 then
		tileBatch[z]:set(y*width+x-1,tileQuads[v],x*tileWidth,y*tileHeight,0,1,1,tileWidth,tileHeight)
	else
		tileBatch[z]:setq(y*width+x-1,tileQuads[v],x*tileWidth,y*tileHeight,0,1,1,tileWidth,tileHeight)
	end
end

local function resetMap()
	for _,tb in pairs(tileBatch) do
		tb:clear()
		for i = 1, length+1 do
			for j = 1, width do
				if version >= 0.9 then
					tb:add(tileQuads[0],j*tileWidth,i*tileHeight,0,1,1,tileWidth,tileHeight)
				else
					tb:addq(tileQuads[0],j*tileWidth,i*tileHeight,0,1,1,tileWidth,tileHeight)
				end
			end
		end
	end
end

local function reset()
	for i,_ in pairs(map) do
		map[i] = nil
	end
	for i,_ in pairs(path) do
		path[i] = nil
	end
	path[1] = math.random(1,width)
	path[2] = math.random(1,length)
	path[3] = math.random(1,height)
	zoom = path[3]
	resetMap()
	collectgarbage('collect')
end

function love.load()
	tileset = love.graphics.newImage('exampleTileset.png')
	for i = 0,63 do
		tileQuads[i] = love.graphics.newQuad(math.mod(i,8)*tileWidth,math.floor(i/8)*tileHeight,tileWidth,tileHeight,tileset:getWidth(),tileset:getHeight())
	end
	for i = 1, height do
		tileBatch[i] = love.graphics.newSpriteBatch(tileset,width*length*height)
	end
	reset()
	updateCaption()
end

function love.keypressed(k)
	if k == 'left' then
		speed=math.max(1,speed-1)
		updateCaption()
	elseif k == 'right' then
		speed=math.min(30,speed+1)
		updateCaption()
	elseif k == 'up' then
		zoom = math.min(height,zoom+1)
		updateCaption()
	elseif k == 'down' then
		zoom = math.max(1,zoom-1)
		updateCaption()
	end
	if string.sub(kc,ck,ck+string.len(k)-1) == k then
		ck=ck+string.len(k)
		if ck > string.len(kc) then
			reset()
			ck=1
		end
	else ck=1 end
end

function love.mousepressed(x,y,b)
	if b == 'wd' then
		zoom = math.max(1,zoom-1)
		updateCaption()
	elseif b == 'wu' then
		zoom = math.min(height,zoom+1)
		updateCaption()
	end
end

local function getMap(x,y,z)
	if map[z] then
		if map[z][y] then
			if map[z][y][x] then
				return map[z][y][x]
			elseif z >= 1 and z <= height and y >= 1 and y <= length and x >= 1 and x <= width then
				return 0
			end
		elseif z >= 1 and z <= height and y >= 1 and y <= length and x >= 1 and x <= width then
			return 0
		end
	elseif z >= 1 and z <= height and y >= 1 and y <= length and x >= 1 and x <= width then
		return 0
	end
end

local function setMap(x,y,z,v)
	map[z] = map[z] or {}
	map[z][y] = map[z][y] or {}
	map[z][y][x] = v
end

local function merge(x1,y1,z1,x2,y2,z2)
	--print(x1,y1,z1,x2,y2,z2)
	--print(getMap(x1,y1,z1),getMap(x2,y2,z2))
	if x1 > x2 then
		setMap(x1,y1,z1,getMap(x1,y1,z1)+left)
		setMap(x2,y2,z2,getMap(x2,y2,z2)+right)
	elseif x1 < x2 then
		setMap(x1,y1,z1,getMap(x1,y1,z1)+right)
		setMap(x2,y2,z2,getMap(x2,y2,z2)+left)
	elseif y1 > y2 then
		setMap(x1,y1,z1,getMap(x1,y1,z1)+front)
		setMap(x2,y2,z2,getMap(x2,y2,z2)+back)
	elseif y1 < y2 then
		setMap(x1,y1,z1,getMap(x1,y1,z1)+back)
		setMap(x2,y2,z2,getMap(x2,y2,z2)+front)
	elseif z1 > z2 then
		setMap(x1,y1,z1,getMap(x1,y1,z1)+down)
		setMap(x2,y2,z2,getMap(x2,y2,z2)+up)
	elseif z1 < z2 then
		setMap(x1,y1,z1,getMap(x1,y1,z1)+up)
		setMap(x2,y2,z2,getMap(x2,y2,z2)+down)
	end
end

local function updateMaze(x,y,z)
	setMaze(x,y,z,getMap(x,y,z))
end

function love.update(dt)
	for i = 1, speed do
		if #path > 0 then
			local t = {}
			local x,y,z = path[#path-2],path[#path-1],path[#path]
			if getMap(x-1,y,z) == 0 then
				t[#t+1]={x-1,y,z}
			end
			if getMap(x+1,y,z) == 0 then
				t[#t+1]={x+1,y,z}
			end
			if getMap(x,y-1,z) == 0 then
				t[#t+1]={x,y-1,z}
			end
			if getMap(x,y+1,z) == 0 then
				t[#t+1]={x,y+1,z}
			end
			if getMap(x,y,z-1) == 0 then
				t[#t+1]={x,y,z-1}
			end
			if getMap(x,y,z+1) == 0 then
				t[#t+1]={x,y,z+1}
			end
			if #t > 0 then
				local c = math.random(1,#t)
				local x1,y1,z1 = unpack(t[c])
				merge(x,y,z,x1,y1,z1)
				path[#path+1]=x1
				path[#path+1]=y1
				path[#path+1]=z1
				updateMaze(x,y,z)
				updateMaze(x1,y1,z1)
			else
				for i = 1, 3 do
					path[#path]=nil
				end
			end
		else
			break
		end
	end
end

function love.draw()
	for i,v in pairs(tileBatch) do
		local a = 0
		if i > zoom then
			break
		elseif i == zoom then
			a = 255
		elseif i == zoom-1 then
			a = 127
		elseif i == zoom-2 then
			a = 63
		end
		love.graphics.setColor(a,a,a,255)
		love.graphics.draw(v,math.mod(800,16)/2,math.mod(600,16)/2)
	end
	love.graphics.setColor(255,0,0,127)
	for i = 3, #path, 3 do
		local x,y,z = path[i-2],path[i-1],path[i]
		if z == zoom then
			love.graphics.rectangle('fill',math.mod(800,16)/2+3+(x-1)*tileWidth,math.mod(600,16)/2+3+(y-1)*tileHeight,tileWidth-6,tileHeight-6)
		end
	end
end
